@@ -1,4 +1,4 @@
-{-# OPTIONS -farrows #-}
+{-# LANGUAGE Arrows #-}
 
 {- 
 ParseSyntaxFiles.hs processes a directory containing Kate
@@ -22,7 +22,6 @@ module Main where
 
 import Text.XML.HXT.Arrow
 import Text.XML.HXT.Arrow.Edit
-import Text.XML.HXT.Arrow.XmlNodeSet
 import Control.Arrow
 import Control.Arrow.ArrowList
 import Data.List
@@ -36,7 +35,6 @@ import System.Exit
 import System.FilePath
 import Text.PrettyPrint
 import Text.Highlighting.Kate.Definitions
-import Text.Highlighting.Kate.Common (capitalize)
 
 data SyntaxDefinition =
   SyntaxDefinition { synLanguage      :: String
@@ -83,25 +81,32 @@ data SyntaxParser =
                } deriving (Read, Show)
 
 
+-- | Converts a list of files (ending in .xml) and directories containing .xml files
+-- into a list of .xml files.
+argFiles :: [String] -> IO [String]
+argFiles [] = error "Specify paths of xml files and/or directories containing xml syntax files."
+argFiles args = do
+  let isXmlFile x = isSuffixOf ".xml" x
+  let (files, dirs) = partition isXmlFile args
+  dirContents <- mapM (\dir -> getDirectoryContents dir >>= return . map (combine dir) . filter isXmlFile) dirs
+  return $ nub (files ++ concat dirContents)
+
+
 libraryPath = joinPath ["Text", "Highlighting", "Kate"]
 destDir = joinPath [libraryPath, "Syntax"]
 
 main :: IO ()
 main = do
-  argv <- getArgs
-  srcDir <- if null argv 
-               then error "Specify path of directory containing xml syntax files."
-               else return $ argv !! 0
-  let isXmlFile x = isSuffixOf ".xml" x
-  files <- getDirectoryContents srcDir >>= (return . map (combine srcDir) . filter isXmlFile)
+  files <- getArgs >>= argFiles
   destDirExists <- doesDirectoryExist destDir
   if destDirExists
      then return ()
      else createDirectory destDir 
-  mapM processOneFile files >> return ()
+  mapM_ processOneFile files
   let syntaxFile = combine libraryPath (addExtension "Syntax" "hs")
   putStrLn $ "Writing " ++ syntaxFile
-  let names = sort $ map nameFromPath files 
+  -- Get all syntax files, not only the newly generated ones.
+  names <- getDirectoryContents destDir >>= return . sort . map dropExtension . filter (isSuffixOf ".hs")
   let imports = unlines $ map (\name -> "import qualified Text.Highlighting.Kate.Syntax." ++ name ++ " as " ++ name) names 
   let cases = unlines $ map (\name -> "        " ++ show (map toLower name) ++ " -> " ++ name ++ ".highlight") names
   let languageExtensions = concat $ intersperse ", " $ map (\name -> "(" ++ show name ++ ", " ++ name ++ ".syntaxExtensions)") names
@@ -134,9 +139,15 @@ main = do
            \            -> String                        -- ^ Source code to highlight\n\
            \            -> Either String [SourceLine]    -- ^ Either error message or result\n\
            \highlightAs lang =\n\
-           \  case (map toLower lang) of\n" ++
+           \  let lang'  = map toLower lang\n\
+           \      lang'' = if lang' `elem` map (map toLower) languages\n\
+           \                  then lang'\n\
+           \                  else case languagesByExtension lang' of\n\
+           \                            [l]  -> map toLower l  -- go by extension if unambiguous\n\
+           \                            _    -> lang'\n\
+           \  in  case lang'' of\n" ++
            cases ++
-           "        _ -> (\\_ -> Left (\"Unknown language ++ \" ++ lang))\n"
+           "        _ -> (\\_ -> Left (\"Unknown language: \" ++ lang))\n"
 
 processOneFile :: FilePath -> IO ()
 processOneFile src = do
@@ -238,12 +249,6 @@ mkParser syntax =
                                       endLineParser, withAttr, styles, parseExpressionInternal, 
                                       defaultAttributes {- , lineBeginContexts -}] ++ contexts ++ [contextCatchAll]
 
-mkIdentifier :: String -> String
-mkIdentifier "" = ""
-mkIdentifier ('-':x:xs) = toUpper x : mkIdentifier xs
-mkIdentifier ('-':xs) = mkIdentifier xs
-mkIdentifier (x:xs) = x : mkIdentifier xs
-
 mkAlternatives :: [Doc] -> Doc
 mkAlternatives docs = 
   let contents = vcat $ intersperse (text "<|>") docs
@@ -338,6 +343,10 @@ langNameToModule str =  "Text.Highlighting.Kate.Syntax." ++
     "JavaScript" -> "Javascript"
     x -> x
 
+capitalize :: String -> String
+capitalize (x:xs) = toUpper x : xs
+capitalize [] = []
+
 nameFromPath :: FilePath -> String
 nameFromPath = concat . map capitalize . words . 
                (map (\c -> if c == '-' then ' ' else c)) . takeFileName . 
